/* This header file contains the keys for the shared memory segments as
 * well as the keys for the semaphores used
 */

#define RESOLUTION 1000000
#define DEFAULT_PERMISSIONS 0644

#define SHM_CONTROLLER_KEY 13380
#define SHM_COUNTER_KEY 13390
#define SHM_DRAM_KEY 135

#define SHM_L2_KEY 20000

#define SEMAPHORE_CONTROLLER_KEY 9987
#define SEMAPHORE_SYNCH_KEY 8901
#define SEMAPHORE_REPORT_KEY 7891
#define SEMAPHORE_L2_KEY 197

/* Commands issued through shared memory */

#define WAITING_FOR_COMMAND 0
#define RUN_COMMAND 1
#define SIMULATION_COMPLETED_COMMAND 2
#define HALT_COMMAND 3

/* Some useful functions for locking */

/* This function simply locks a semaphore */
void my_lock(int semaphore_id, int semaphore_number);

/* This function unlocks a semaphore */
void my_unlock(int semaphore_id, int semaphore_number);

/* This function destroys a semaphore */
void destroy_semaphore(int semaphore_id);

/* This function creates a new semaphore set and returns the id */
int create_semaphore_set(key_t semaphore_key, int setsize);

/* This function gets a semaphore set based on the key */
int get_semaphore_set(key_t semaphore_key, int setsize);

/* This function sets the value of a semaphore (useful for initialization */
void semaphore_set_value(int semaphore_id, int semaphore_number, int value);

/* Shared memory operations */
/* Create a shared memory segment based on a key with a set size*/
int create_shmem(key_t key, int size);

/* Get a shared memory segment based on a key with a set size*/
int get_shmem(key_t key, int size);

/* This function returns a pointer to the shared memory segment */
void *shmem_attatch(int shmem_id);
