/*
 * performance.c
 *
 * The purpose of this small program is to show how userspace can access
 * performance registers. This code is pretty useless unless RDPMC is enabled
 * for userspace. See the kernel module for additional details.
 *
 * Written by Marius Grannas 2006
 *
 * Note: Some of this code is athlon XP specific and might break on other 
 * architectures.
 *
 * Compile with
 * gcc performance.c -o performance
 *
 * Note: Do not turn on optimizing, the compiler might optimize away the entire
 * loop body!
 */

#include <stdio.h>
#include <stdlib.h>

/* 
 * This function reads the value of the built in clock cycle timer 
 * The cpuid instruction is there to ensure serializability (eg it forces
 * a pipeline flush
 */
static inline unsigned long long rdtsctime() {
    unsigned int eax, edx;
    unsigned long long val;
    __asm__ __volatile__("cpuid": : : "ax", "bx", "cx", "dx");
    __asm__ __volatile__("rdtsc":"=a"(eax), "=d"(edx));
    val = edx;
    val = val << 32;
    val += eax;
    return val;
}

/* 
 * This function reads the value of performance register 0.
 * Cpuid instruction is again used for serializability
 */
static inline unsigned long long readpc() {
  unsigned int eax, edx;
  unsigned long long val;
  __asm__ __volatile__("cpuid": : : "ax", "bx", "cx", "dx");
  __asm__ __volatile__("xor %%ecx, %%ecx ; rdpmc"
                        :"=a"(eax), "=d"(edx) /* Output */
                        :                     /* Input  */
                        :"ecx"                /* Clobbered */ 
                        );
  val = edx;
  val = val << 32;
  val += eax;
  return val;
}

/*
 * Sample test program.
 * This shows how the counter works as well as the timer.
 */

int main(void) {
  long long int start,stop;
  int a = 9;
  int c = 3;
  int d[90000];
  int i;
  
  printf("Clock is %lld\n", rdtsctime());
  start = readpc();
  
  for (i=0; i< 1000; i++) {
  	d[i] = a / c;
  }

  stop = readpc();
  printf("Number of misses: %lld\n", stop-start);
  printf("Clock is %lld\n", rdtsctime());
  return 0;
}
