/*  
 *  pcm.c 
 *
 *  This linux kernel module enables RDPMC for user level programs.
 *  In addition, it sets up counter #0 to count the number of cache misses.
 *
 *  Written by Marius Grannas 2006
 *
 * NOTE! This is AMD Athlon XP specific code, and will possibly break things
 * badly on any other arch.
 * 
 */

#include <linux/module.h> /* Needed by all modules */
#include <linux/kernel.h> /* Needed for KERN_INFO */

/* This function is called when the module is loaded */
int init_module(void)
{
  printk(KERN_INFO "Enabling RDPCM...\n");
  
  // The following is a reimplementation of set_in_cr4() that doesn't
  // require mmu_cr4_features which isn't exported anymore.

  __asm__ __volatile__("movl %%cr4,%%eax\n\t"
                       "orl %0,%%eax\n\t"
                       "movl %%eax,%%cr4\n"
                       : : "irg" (X86_CR4_PCE)
                       :"ax");

  printk(KERN_INFO "Setting up counter 0.\n");
  // This part sets up counter 0 to count the number of data cache misses.
  // See the Athlon Optimization guidelines for values
  __asm__ __volatile__ ("mov  $0xC0010000, %%ecx ; mov $0x00420041, %%eax; wrmsr"
		       :
		       :
		       : "eax", "ecx");

  // A non 0 return means init_module failed; module can't be loaded. 
  return 0;
}


/* This function is called when the module is unloaded */

void cleanup_module(void)
{
  printk(KERN_INFO "Disabling Performance counters.\n");
  // Same as above
  __asm__ __volatile__("movl %%cr4,%%eax\n\t"
                       "andl %0,%%eax\n\t"
                       "movl %%eax,%%cr4\n"
                       : : "irg" (~X86_CR4_PCE)
                       :"ax");

}
